print("import: 'transonic'")
import transonic

print("import: 'transonic.analyses'")
import transonic.analyses

print("import: 'transonic.backends'")
import transonic.backends

print("import: 'transonic_cl'")
import transonic_cl

